;**************************************************************
;***      Universal Light Controller:                       ***
;***      Routines for evaluation of RC pulses              ***
;**************************************************************

;History:
;1.0 initial Version
;1.0 added support of PIC16F648A
;1.1 added RC inactive switch (#48 = MSB(rcdetect)   


;========================== compute RC-values from received pulses
;(TMR1 contains pulse width, unit: 4s)
#define RCADJ 22		;value to adjust for middle = 64
proc_rc
 if RCCHAN >1			;two rc-inputs:
	call	rcswap		;if ch#2: swap ch2-data to lower nibble
	call	fsrrcval	;FSR = &rcval(ch)
	movf	INDF,W		;current value for processing of memory switch
 else
	movf	rcval,W		;current value for processing of memory switch
 endif
	call	rcswpos		;get current switch-status
	movwf	temp		;save...

	movlw	200		;subtract min. pulse with = 0.8 ms from timer
	subwf	TMR1L,F
	btfss	STATUS,C
	decf	TMR1H,F
	
	btfsc	TMR1H,7		;negative result?
	goto	outrange	;y: pulse is too short

	bcf	STATUS,C
	rrf	TMR1H,W		;/2: time division = 8s
	andlw	0xFF
	btfss	STATUS,Z
	goto	outrange	;pulse is too long
	rrf	TMR1L,W		;pw-800 -> W (8s units)
	movwf	y		;1..2 ms = 25..150
	movwf	x		;save new rc pulse width

	movf	rclast,W
 if RCCHAN >1
	btfsc	flags,RCSELECT
	movf	rclast+1,W	
 endif	
	subwf	y,W		;compare previous with current value for error det.
	movwf	y		
	btfsc	y,7	
	xorlw	0xFF		;neg.-> invert
	andlw	0xFC		;Z=0 if large difference to previous pulse
	btfsc	STATUS,Z
	goto	rcvalok		;value ok.

	bsf	errcnt,0	;pulse very different from last pulse,
	goto	rcskip		;maybe it's error

;Current pulse seems to be ok. 
;If there was an error before (errcnt != 0): need some consecutive right pulses 
rcvalok
	movlw	4		;reset detection timer = 0.5s
	movwf	rcdetect
	
	movlw	0x0F
	andwf	errcnt,W
	btfsc	STATUS,Z
	goto	rcnewval	;errcnt == 0: pulse is valid

	decf	errcnt,F	;just wait for more valid pulses
	goto	rcskip

;received pulse width is verified ok!	
rcnewval
 if RCCHAN >1
	movf	INDF,W		;get difference from last valid RC-value...
 else
	movf	rcval,W		
 endif
	addlw	RCADJ		;=> "raw" value of current rcval
	subwf	x,W		;new-old val
	btfsc	STATUS,C
	goto	rcnvge		;new val >= old val 
;new val < old val:
	addlw	RCHYST
	btfsc	STATUS,C
	goto	rcskip		;difference < hysteresis: keep old value

	movlw	RCADJ-RCHYST
	goto	rcnewval1	;new val = recvd.val + hysteresis

;new value >= old value:
rcnvge
	addlw	-RCHYST
	btfss	STATUS,C
	goto	rcskip		;difference < hysteresis: keep old value

	movlw	RCADJ+RCHYST	;new val = recvd.val - hysteresis
rcnewval1
	subwf	x,W		;adj. for middle pos. = 64 (+/- hysteresis)
	btfss	STATUS,C
	movlw	0		;<0 -> set to 0

 if RCCHAN >1
;____________________________ 2 RC channels: ______________
	movwf	INDF		;store new val (->rcval) 
	call	rcswpos		;get new status of switch
	subwf	temp,W		
	btfss	STATUS,Z	;any change from last status?
	call	domemsw		;y: process memory switch 
rcskip
	movf	x,W
	btfss	flags,RCSELECT
	movwf	rclast		;ref. value for rc-channel 1
	btfsc	flags,RCSELECT
	movwf	rclast+1	;ref. value for rc-channel 2
	goto	rcready

outrange
	bsf	errcnt,1
	
rcready
	call	rcswap		;if ch#2: swap RC2-data back to higher nibble

;change rc channel:
	btfss	flags,RC2ENA
	goto	procrcx		;other channel is not active
rcchch
	btfsc	flags,RCSELECT	;was ch2 selected?
	goto	rcsel1		;y: select ch1

	bsf	flags,RCSELECT	;sel. ch2
	bsf	STATUS,RP0
 errorlevel-302
 ifdef __12F675
	bcf	IOC,3		;disable int. for RC ch 1
	bsf	IOC,5		;interrupt on change for RC ch2
 else
	bcf	IOCA,3		;disable port change irq ch1
	bsf	IOCA,0		;enable port change irq ch2
 endif
 errorlevel+302
	bcf	STATUS,RP0
	goto	procrcx

rcsel1	
	bcf	flags,RCSELECT	;-> ch1
	bsf	STATUS,RP0
 errorlevel-302
 ifdef __12F675
	bcf	IOC,5		;disable int. for RC ch 2
	bsf	IOC,3		;interrupt on change for RC ch1
 else
	bcf	IOCA,0		;disable port change irq ch2
	bsf	IOCA,3		;enable port change irq ch1
 endif
	bcf	STATUS,RP0
 errorlevel+302

 else
;____________________________ 1 RC channel: ______________

	movwf	rcval		;new rc value
	call	rcswpos		;get new status of switch
	subwf	temp,W		
	btfss	STATUS,Z	;any change from last status?
	call	domemsw		;y: process memory switch 
rcskip
	movf	x,W
	movwf	rclast		;new received value
	goto	procrcx

outrange
	bsf	errcnt,0
	
 endif
;__________________________________________________________

procrcx
	clrf	TMR1L		;clr timer (will start by rc-interrupt)
	clrf	TMR1H
 ifdef __16F648A
	bcf	INTCON,INTF
	bsf	INTCON,INTE	;enable RC-interrupt
 else
	bsf	INTCON,RAIE	;enable port change int.
 endif
	return

;========= status of rc switch changed: process memory switch
;temp = previous status of switch (-1/0/1) 
domemsw
	btfss	flags,RCINIT
	return			;rc not yet fully initialized

	movf	temp,W		
	btfsc	STATUS,Z	;old pos = 0 (middle)?
	goto	doms_act	;y: change 0->1 or 0-> -1	

;switch was released from active position:
	movf	rctimer,W
	andlw	0x0F		;mask bits of this channel
	addlw	-MS_LONG
	btfsc	STATUS,C
	goto	doms_act	;rel. after long time, already processed!

;switch released after short time: toggle primary ms-bit
	movlw	B'0001'		;if sw was -1: toggle bit 0
	btfss	temp,7	
	movlw	B'0010'		;sw was +1: toggle bit 1		 
	xorwf	memswtch,F 

doms_act
 if RCCHAN>1
	movf	INDF,W
 else
	movf	rcval,W
 endif
	call	rcswpos		;W= -1/0/1 (new position of switch)
	andlw	0xFF
	btfsc	STATUS,Z
	return			;new pos.= middle: done!

	movlw	0xF0
	andwf	rctimer,F	;start timer
	return
	
;========= 1/10s RC service during initialization
;finds out, which RC inputs receive rc pulses
rc_init
 if RCCHAN>1
	incf	rctimer,F
	btfss	rctimer,2	
	return			;0.4s not yet over, exit

	clrf	rctimer
	call	fsrrcval	;set FSR=rcval(old channel)
	call	rcchch		;read other channel next
	incfsz	INDF,W		;got pulses on old channel?
	goto	rc_init1	;y: read other channel or init.complete

;no pulses on just selected channel:
	call	fsrrcval	;FSR = rcval(new channel)
	incfsz	INDF,W		;got pulses already here?
	bsf	flags,RCINIT	;y: this is the only active channel
	return

;got pulses on just selected ch.:
rc_init1
	call	fsrrcval	;FSR = rcval(new channel)
	incfsz	INDF,W		;already got pulses here?
	goto	rc_ena2ch	;y: both channels are active!
	return			;n: read pulses from new channel

rc_ena2ch
	bsf	flags,RCINIT	;yes: rc initialization complete
	bsf	flags,RC2ENA	;two channels are active
	return

 else
	incfsz	rcval,W		;wait for valid rc pulses
	bsf	flags,RCINIT	;got them: init complete!
	return
 endif

;========= service routine, called every 1/10th sec.
rcserv10
	btfss	rcdetect,7	;decrement if >= 0
	decf	rcdetect,F	;runs to -1 if no RC pulses are detected

	btfss	flags,RCINIT
	goto	rc_init		;initialization phase

 if RCCHAN >1
	movlw	rcval+1
	movwf	FSR		;2nd channel first
	call	rcserv10a	;process ch 2
	decf	FSR,F		;then process ch 1

rcserv10a
	call	rcswap1		;data -> lower nibbles

	incf	INDF,W		;0xFF->0 if rc channel is inactive
	btfsc	STATUS,Z
	return			;rc channel is not active
 endif
	movf	rctimer,W
	andlw	0x0F
	addlw	-MS_LONG
	btfsc	STATUS,C
	return			;time was already over - nothing more to do here!
	incf	rctimer,F
	addlw	1		
	btfss	STATUS,Z	;reached time MS_LONG after increment?
	return			;not yet

;timeout (MS_LONG):
 if RCCHAN >1
	movf	INDF,W		;rcval(ch)->W 
 else
	movf	rcval,W
 endif
	call	rcswpos
	andlw	0xFF
	btfsc	STATUS,Z
	return			;switch in middle pos.-> exit

	addlw	1		;C=1 if switch pos = -1
	movlw	B'0100'		;if sw = -1: toggle bit 2
	btfss	STATUS,C
	movlw	B'1000'		;sw = +1: toggle bit 3
	xorwf	memswtch,F
	return

 if RCCHAN >1
;========== Swaps RC parameter, if channel 2 is selected
rcswap
	btfss	flags,RCSELECT	;ch#2 selected?
	return			;no
rcswap1
	swapf	errcnt,F	
	swapf	rctimer,F
	swapf	memswtch,F
	return

;========== sets FSR = rcval for selected channel
fsrrcval
	movlw	rcval
	btfsc	flags,RCSELECT	;ch#2 selected?
	movlw	rcval+1		;y
	movwf	FSR
	return
 endif

;========= returns switch-position (-1/0/1) from value (W)
rcswpos
	addlw	-RCPW_LOW
	btfss	STATUS,C
	retlw 	-1		;<RCPW_LOW
	addlw	-(RCPW_HIGH-RCPW_LOW)
	btfss	STATUS,C
	retlw 	0		;<RCPW_HIGH
	retlw	1		;>= RC_PWHIGH

;**************************************************************
;RC Interrupt Service Routine
;**************************************************************

;RC input level changed: start/stop timer
rc_isr
 ifdef __16F648A
	btfss	INTCON,INTE	;rc interrupt disabled?
	goto	t0_isr		;y: must be T0-irq!

	bcf	flags,SAVERP0 
	btfsc	STATUS,RP0	;save RP0-bit...
	bsf	flags,SAVERP0
	bcf	STATUS,RP0	;select bank 0

	btfss	PORTB,0		;falling edge?
	goto	rcisr_stop	;y: end of pulse	

rcisr_start
	bcf	PIR1,TMR1IF
	bsf	T1CON,TMR1ON	;start timer
 errorlevel -302
	bsf	STATUS,RP0
	bcf	OPTION_REG,INTEDG ;next int on falling edge
	goto	rcisr_exit

rcisr_stop
	bcf	T1CON,TMR1ON	;measuring complete, stop timer
	bcf	INTCON,INTE	;main prg will set INTE after processing value
	bsf	STATUS,RP0
	bsf	OPTION_REG,INTEDG ;next int on rising edge

rcisr_exit
 errorlevel +302
	bcf	STATUS,RP0
	btfsc	flags,SAVERP0
	bsf	STATUS,RP0	;restore RB0
	bcf	INTCON,INTF	;clear RC interrupt flag
	retfie

 else
	btfss	INTCON,RAIE	;rc interrupt disabled?
	goto	t0_isr		;y: must be T0-irq!

	bcf	flags,SAVERP0 
	btfsc	STATUS,RP0	;save RP0-bit...
	bsf	flags,SAVERP0
	bcf	STATUS,RP0	;select bank 0

	btfsc	T1CON,TMR1ON	;timer 1 running?
	goto	rcisr_stop	;y: must be end of pulse	
   if RCCHAN >1
	btfsc	flags,RCSELECT	;which channel?
	goto	rcisr_ch2	;ch.2
   endif
   ifdef __12F675
	btfss	GPIO,3
   else
	btfss	PORTA,3
   endif
	goto	rcisr_exit	;H->L transition, don't start timer
	goto	rcisr_start	;start timer

   if RCCHAN >1
rcisr_ch2
     ifdef __12F675
	btfss	GPIO,5
     else
	btfss	PORTA,0
     endif
	goto	rcisr_exit	;H->L transition, don't start timer
   endif

rcisr_start
	bcf	PIR1,TMR1IF
	bsf	T1CON,TMR1ON	;start timer
	goto	rcisr_exit

rcisr_stop
	bcf	T1CON,TMR1ON	;measuring complete, stop timer
	bcf	INTCON,RAIE	;main prg will set RAIE after processing value

rcisr_exit
   ifdef __12F675
	bcf	GPIO,3		;(dummy op.) clr mismatch condition 
   else
	bcf	PORTA,3		;(dummy op.) clr mismatch condition 
   endif
	btfsc	flags,SAVERP0
	bsf	STATUS,RP0	;restore RB0
	bcf	INTCON,RAIF	;clr interrupt flag

	retfie
 endif
